"""Rio-Cogeo Errors and Warnings."""


class DeprecationWarning(UserWarning):
    """Rio-cogeo module deprecations warning."""


class LossyCompression(UserWarning):
    """Rio-cogeo module Lossy compression warning."""


class IncompatibleBlockRasterSize(UserWarning):
    """Rio-cogeo module incompatible raster block/size warning."""
